#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import base64
import inspect
import json
from typing import Any, Type, get_type_hints

from mesh.macro.ark import T
from mesh.macro.compatible import Compatible


class Serializable(object):
    """ Serializable """

    def __init__(self, cls=None):
        self.cls = cls
        annotations = getattr(cls, '__annotations__') if hasattr(cls, '__annotations__') else {}
        for attr in inspect.classify_class_attrs(cls):
            if not attr.name or attr.name.startswith("_") or not attr.object or not hasattr(attr.object, 'kind'):
                continue
            annotations[attr.name] = getattr(attr.object, 'kind')
        setattr(cls, 'decode', self.decode)
        setattr(cls, 'encode', self.serialize)
        setattr(cls, '__annotations__', annotations)

    def serialize(self, value: Any) -> Any:
        if isinstance(value, bytes):
            return base64.b64encode(value).decode("utf-8")

        if isinstance(value, (int, float)):
            return value

        return vars(value)

    def deserialize(self, instance: Any, **kwargs) -> Any:
        if kwargs is None:
            return instance
        types = get_type_hints(instance.__class__)
        if types and types.__len__() > 0:
            for name, kind in types.items():
                value = kwargs.get(name, None)
                if value is None:
                    continue
                setattr(instance, name, self.construct(value, kind))

            return instance

        members = inspect.getmembers(instance)
        for name, _ in members:
            value = kwargs.get(name, None)
            if value is None or not name or name.startswith('_'):
                continue
            setattr(instance, name, self.construct(value, type(value)))

        return instance

    def construct(self, value: Any, vtp: type) -> Any:
        kind = Compatible.get_origin(vtp) if Compatible.get_origin(vtp) else vtp
        if kind is bool:
            return value if value else False
        if kind is complex:
            return value
        if kind is int or kind is float:
            return value if value else 0
        if kind is str:
            return value if value else ''
        if kind is list:
            sut = dict if Compatible.get_args(vtp).__len__() < 1 else Compatible.get_args(vtp)[0]
            return [self.construct(v, sut) for v in value]
        if kind is tuple:
            sut = dict if Compatible.get_args(vtp).__len__() < 1 else Compatible.get_args(vtp)[0]
            return tuple([self.construct(v, sut) for v in value])
        if kind is set:
            sut = dict if Compatible.get_args(vtp).__len__() < 1 else Compatible.get_args(vtp)[0]
            return {self.construct(v, sut) for v in value}
        if kind is iter:
            sut = dict if Compatible.get_args(vtp).__len__() < 1 else Compatible.get_args(vtp)[0]
            return iter([self.construct(v, sut) for v in value])
        if kind is dict:
            return value
        if kind is bytes:
            if isinstance(value, bytes):
                return value
            return base64.b64decode(value)
        if isinstance(value, (bool, complex, int, float, str,)):
            return value
        sub = kind()
        return self.deserialize(sub, **value)

    def decode(self, text: str):
        struct = json.loads(text)
        if struct is None:
            return None
        instance = self.cls()
        return self.deserialize(instance, **struct)


def serializable(cls: Type[T] = None) -> Type[T]:
    if cls is not None:
        Serializable(cls)
        return cls

    def classic(icls: Type[T] = None) -> Type[T]:
        Serializable(icls)
        return icls

    return classic



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from __future__ import absolute_import, division, unicode_literals

from mesh.codec.codec import Codec
from mesh.codec.jsons import JsonCodec, Json
from mesh.codec.protobuf import ProtobufCodec, Protobuf
from mesh.codec.xml import Xml, XmlCodec
from mesh.codec.yml import YamlCodec, Yaml

__all__ = (
    "Codec", "JsonCodec", "YamlCodec", "ProtobufCodec", "XmlCodec", "Json", "Yaml", "Xml", "Protobuf"
)


def init():
    """ init function """
    pass



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import mesh.log as log
import threading
from abc import ABC, abstractmethod
from enum import Enum
from mesh.kinds.location import Location, Principal
from mesh.macro import T
from typing import Any, Dict, Generic, Deque, Optional, List


class RunMode(Enum):
    ROUTINE = 1
    """
    正常模式
    """
    PERFORM = 2
    """
    评测模式
    """
    DEFENSE = 4
    """
    高防模式
    """
    DEBUG = 8
    """
    调试模式
    """
    LOAD_TEST = 16
    """
    压测模式
    """
    MOCK = 32
    """
    Mock模式
    """

    def is_debug(self) -> bool:
        return RunMode.matches(RunMode.DEBUG, self.value)

    def is_load_test(self) -> bool:
        return RunMode.matches(RunMode.LOAD_TEST, self.value)

    def is_routine(self) -> bool:
        return RunMode.matches(RunMode.ROUTINE, self.value)

    def is_perform(self) -> bool:
        return RunMode.matches(RunMode.PERFORM, self.value)

    def is_defense(self) -> bool:
        return RunMode.matches(RunMode.DEFENSE, self.value)

    def is_mock(self) -> bool:
        return RunMode.matches(RunMode.MOCK, self.value)

    @staticmethod
    def matches(mode: Any, code: int):
        return isinstance(mode, RunMode) and (mode.value & code) is code

    @staticmethod
    def get_by_code(code: int) -> "RunMode":
        for e in RunMode:
            if (e.value & code) is e.value:
                return e
        return RunMode.ROUTINE


class Key(Generic[T], ABC):
    """
    Context key.
    """

    def __init__(self, name: str):
        self.name = name

    @abstractmethod
    def get_if_absent(self) -> T:
        """ Get the attribute if present by key """
        pass

    @abstractmethod
    def map(self, fn) -> Optional[Any]:
        """ Map new value """
        pass

    @abstractmethod
    def if_present(self, fn):
        """ Consume attribute if present """
        pass

    @abstractmethod
    def or_else(self, v: T) -> T:
        """ Get attribute default if optional """
        pass

    @abstractmethod
    def is_present(self) -> bool:
        """ Check the attribute is present """
        pass


class Metadata(Enum):
    """
    https://www.rfc-editor.org/rfc/rfc7540#section-8.1.2
    """

    MESH_SPAN_ID = "mesh-span-id"
    MESH_TIMESTAMP = "mesh-timestamp"
    MESH_RUN_MODE = "mesh-run-mode"
    MESH_CONSUMER = "mesh-consumer"
    MESH_PROVIDER = "mesh-provider"
    MESH_URN = "mesh-urn"
    #
    MESH_INCOMING_HOST = "mesh-incoming-host"
    MESH_OUTGOING_HOST = "mesh-outgoing-host"
    MESH_INCOMING_PROXY = "mesh-incoming-proxy"
    MESH_OUTGOING_PROXY = "mesh-outgoing-proxy"
    MESH_SUBSET = "mesh-subset"
    # PTP
    MESH_VERSION = "x-ptp-version"
    MESH_TECH_PROVIDER_CODE = "x-ptp-tech-provider-code"
    MESH_TRACE_ID = "x-ptp-trace-id"
    MESH_TOKEN = "x-ptp-token"
    MESH_URI = "x-ptp-uri"
    MESH_FROM_NODE_ID = "x-ptp-source-node-id"
    MESH_FROM_INST_ID = "x-ptp-source-inst-id"
    MESH_TARGET_NODE_ID = "x-ptp-target-node-id"
    MESH_TARGET_INST_ID = "x-ptp-target-inst-id"
    MESH_SESSION_ID = "x-ptp-session-id"
    MESH_TOPIC = "x-ptp-topic"
    MESH_TIMEOUT = "x-ptp-timeout"

    def key(self) -> str:
        return self.value

    def set(self, attachments: Dict[str, str], v: str):
        if attachments and '' != v and v != attachments.get(self.key(), ''):
            attachments[self.key()] = v

    def get(self, attachments: Dict[str, str]) -> str:
        return attachments.get(self.key(), '')

    def append(self, attachments: List[Any], v: str):
        if attachments is not None and '' != v:
            attachments.append((self.key(), v))


class Context(ABC):
    """
    MPC invoke context.
    """

    @abstractmethod
    def get_trace_id(self) -> str:
        """
        Get the request trace id.
        :return: The request trace id.
        """
        pass

    @abstractmethod
    def get_span_id(self) -> str:
        """
        Get the request span id.
        :return: The request span id.
        """
        pass

    @abstractmethod
    def get_timestamp(self) -> int:
        """
        Get the request create time.
        :return: The request create time.
        """
        pass

    @abstractmethod
    def get_run_mode(self) -> RunMode:
        """
        Get the request run mode. RunMode
        :return: The request run mode.
        """
        pass

    @abstractmethod
    def get_urn(self) -> str:
        """
        Mesh resource uniform name. Like: create.tenant.omega.json.http2.lx000001.mpi.trustbe.net
        :return: Uniform name.
        """
        pass

    @abstractmethod
    def get_consumer(self) -> Location:
        """
        Get the consumer network principal.
        :return: Consumer network principal.
        """
        pass

    @abstractmethod
    def get_provider(self) -> Location:
        """
        Get the provider network principal.
        :return: Provider network principal.
        """
        pass

    @abstractmethod
    def get_attachments(self) -> Dict[str, str]:
        """
        Dispatch attachments.
        :return: Dispatch attachments.
        """
        pass

    @abstractmethod
    def get_principals(self) -> "Queue[Principal]":
        """
        Get the mpc broadcast network principals.
        :return: Broadcast principals.
        """
        pass

    @abstractmethod
    def get_attributes(self) -> Dict[str, Any]:
        """
        Get the context attributes. The attributes don't be serialized.
        :return: attributes
        """
        pass

    @abstractmethod
    def get_attribute(self, key: Key[T]) -> T:
        """
        Like getAttachments, but attribute wont be transfer in invoke chain.
        :param key: attribute key
        :return: attribute value
        """
        pass

    @abstractmethod
    def set_attribute(self, key: Key[T], value: T) -> None:
        """
        Like putAttachments, but attribute won't be transfer in invoke chain.
        :param key: attribute key
        :param value: attribute
        :return: None
        """
        pass

    @abstractmethod
    def rewrite_urn(self, urn: str) -> None:
        """
        Rewrite the urn.
        :param urn: urn
        :return:
        """
        pass

    @abstractmethod
    def rewrite_context(self, another: "Context") -> None:
        """
        Rewrite the context by another context.
        :param another: another context
        :return:
        """
        pass

    @abstractmethod
    def resume(self) -> "Context":
        """
        Open a new context.
        :return:
        """
        pass

    @abstractmethod
    def encode(self) -> Dict[str, str]:
        """
        Encode context to string.
        :return:
        """
        pass

    @abstractmethod
    def decode(self, attachments: Dict[str, str]) -> None:
        """
        Decode context from string.
        :return:
        """
        pass


class Queue(Deque, Generic[T]):
    def pop(self) -> T:
        if self.__len__() < 1:
            log.warn(f"{threading.currentThread().ident}({threading.currentThread().name}) pop empty queue.")
            return None
        return super().pop()

    def peek(self) -> T:
        if self.__len__() < 1:
            return None
        return self[-1]



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import inspect
from typing import Any, Callable, Generic

from mesh.macro.ark import T


class Index(Generic[T], object):

    def __init__(self, fn=-1, idx=-1, name='', dft='', doc=None, kind=Any, value=None, initializer=None):
        """
        Emulate PyProperty_Type() in Objects/descrobject.c
        Index for protobuf or thrift etc.
        :param fn: decorate function
        :param idx: Index position.
        """
        self.idx = idx
        self.name = name
        self.dft = dft
        self.doc = doc
        self.kind = kind
        self.value = value
        self.initializer = initializer
        if self.idx != -1:
            return
        if fn == -1:
            return
        if type(fn) is int:
            self.idx = fn

    def __call__(self, *args, **kwargs) -> Callable[..., T]:
        self.initializer = args[0]
        if hasattr(self.initializer, "__name__"):
            self.name = self.initializer.__name__
        if hasattr(self.initializer, "__doc__"):
            self.doc = self.initializer.__doc__
        if hasattr(self.initializer, "__annotations__"):
            signature = inspect.signature(self.initializer)
            self.kind = signature.return_annotation

        return self

    def __get__(self, ref, kind) -> T:
        if ref is None:
            return None
        value = ref.__dict__.get(self.name, None)
        if value is None and callable(self.initializer):
            value = self.initializer(ref)
            ref.__dict__[self.name] = value
        return value

    def __set__(self, ref, value: T):
        if ref is None:
            return
        ref.__dict__[self.name] = value

    def __delete__(self, ref):
        if ref is None:
            return
        if ref.__dict__.get(self.name):
            del ref.__dict__[self.name]


def index(fn=-1, *, idx=-1, name='', dft=None) -> Callable[[Callable[..., T]], T]:
    """
    Index for protobuf or thrift/protobuf etc.
    """
    return Index[T](fn, idx, name, dft)



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import ABC, abstractmethod
from typing import Type, Any

from mesh.kinds import Entity
from mesh.macro import mpi, spi, T


@spi("mesh")
class KV(ABC):

    @abstractmethod
    @mpi("mesh.kv.get")
    def get(self, key: str) -> Entity:
        """
        Get the value from kv store.
        :param key:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.kv.put")
    def put(self, key: str, value: Entity):
        """
        Put the value to kv store.
        :param key:
        :param value:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.kv.remove")
    def remove(self):
        """
        Remove the kv store.
        :return:
        """
        pass

    def get_with_type(self, key: str, kind: Type[T]) -> T:
        """
        Get by codec.
        :param key:
        :param kind:
        :return:
        """
        entity = self.get(key)
        return entity.try_read_object(kind) if entity else None

    def put_object(self, key: str, value: Any):
        """
        Put by codec.
        :param key:
        :param value:
        :return:
        """
        self.put(key, Entity.wrap(value))



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import List

from mesh.context import MeshKey
from mesh.kinds import Route, Environ, Versions, Paging, Page, Institution
from mesh.macro import spi
from mesh.mpc import ServiceProxy, Mesh
from mesh.prsim import Network


@spi("mesh")
class MeshNetwork(Network):

    def __init__(self):
        self.proxy = ServiceProxy.default_proxy(Network)
        self.environ: Environ = None
        self.environ_key = MeshKey("mesh-environ", Environ)

    def get_environ(self) -> Environ:
        if self.environ:
            return self.environ
        if Mesh.context().get_attribute(self.environ_key):
            return Mesh.context().get_attribute(self.environ_key)
        environ = Environ()
        environ.node_id = "LX0000000000000"
        environ.inst_id = "JG0000000000000000"
        Mesh.context().set_attribute(self.environ_key, environ)

        # self.environ = Mesh.context_safe(lambda: self.get_environ_safe())
        # return self.environ
        return environ

    def get_environ_safe(self) -> Environ:
        Mesh.context().get_principals().clear()
        return self.proxy.get_environ()

    def accessible(self, route: Route) -> bool:
        return self.proxy.accessible(route)

    def refresh(self, routes: List[Route]):
        return self.proxy.refresh(routes)

    def get_route(self, node_id: str) -> Route:
        return self.proxy.get_route(node_id)

    def get_routes(self) -> List[Route]:
        return self.proxy.get_routes()

    def get_domains(self) -> List[Route]:
        return self.proxy.get_domains()

    def put_domains(self, domains: List[Route]):
        return self.proxy.put_domains(domains)

    def weave(self, route: Route) -> None:
        return self.proxy.weave(route)

    def ack(self, route: Route) -> None:
        return self.proxy.ack(route)

    def disable(self, node_id: str) -> None:
        return self.proxy.disable(node_id)

    def enable(self, node_id: str) -> None:
        return self.proxy.enable(node_id)

    def index(self, index: Paging) -> Page[Route]:
        return self.proxy.index(index)

    def version(self, node_id: str) -> Versions:
        return self.proxy.version(node_id)

    def instx(self, index: Paging) -> Page[Institution]:
        return self.proxy.instx(index)

    def instr(self, institutions: List[Institution]) -> str:
        return self.proxy.instr(institutions)

    def ally(self, node_ids: List[str]) -> None:
        self.proxy.ally(node_ids)

    def disband(self, node_ids: List[str]) -> None:
        self.proxy.disband(node_ids)



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import unittest

from mesh.metrics.collector import Collector


class TestCollector(unittest.TestCase):

    def test_collect(self):
        Collector().collect()


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from abc import abstractmethod
from typing import Any

from mesh.ioc import Context


class EnvsProcessor:

    @abstractmethod
    def post_properties(self, context: Context) -> Any:
        pass



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#


from mesh.boost.disruptor import Disruptor, BatchConsumer, Consumer, RingBuffer, RingBufferLagStats
from mesh.boost.mooter import Mooter
from mesh.boost.runhook import Runtime
from mesh.boost.scheduler import PythonScheduler

__all__ = (
    "Disruptor", "BatchConsumer", "Consumer", "RingBuffer", "RingBufferLagStats", "Mooter", "Runtime",
    "PythonScheduler")


def init():
    """ init function """
    pass



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

# https://github.com/pypa/pip/issues/3969
# or False if you want to disable it
manylinux1_compatible = True



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#


import inspect
from abc import abstractmethod
from typing import Any, Generic, Type, List, Dict, Callable

from mesh.macro.ark import T


class InvocationHandler:
    """
    InvocationHandler is the interface implemented by the invocation handler of a proxy instance.
    Each proxy instance has an associated invocation handler. When a method is invoked on a proxy instance,
    the method invocation is encoded and dispatched to the invoke method of its invocation handler.
    """

    @abstractmethod
    def invoke(self, proxy: Any, method: Any, *args, **kwargs):
        pass


class InvocationException(Exception):
    """
    Invocation exceptions.
    """

    def __init__(self, cls):
        super(InvocationException, self).__init__(cls, 'is not a InvocationHandler')


class Proxy(Generic[T]):
    """
    Abstract class dynamic proxy
    https://code.activestate.com/recipes/496741-object-proxying/

    Proxy provides static methods for creating objects that act like instances of interfaces but allow for customized
    method invocation. To create a proxy instance for some interface Foo:

    handler = new MyInvocationHandler(...)
    foo = (Foo) Proxy(cls, handler)

    A proxy class is a class created at runtime that implements a specified list of interfaces, known as
    proxy interfaces. A proxy instance is an instance of a proxy class. Each proxy instance has an
    associated invocation handler object, which implements the interface InvocationHandler. A method invocation
    on a proxy instance through one of its proxy interfaces will be dispatched to the invoke method of the instance's
    invocation handler, passing the proxy instance, a java.lang.reflect.Method object identifying the method that was
    invoked, and an array of type Object containing the arguments. The invocation handler processes the encoded method
    invocation as appropriate and the result that it returns will be returned as the result of the method invocation
    on the proxy instance.
    """

    __slots__ = ["_target_", "__weakref__"]

    def __target__(self):
        return object.__getattribute__(self, "_target_")

    def __handler__(self):
        return object.__getattribute__(self, "_handler_")

    def __nonzero__(self):
        return bool(self.__target__())

    def __init__(self, target: Any, handler: InvocationHandler):
        object.__setattr__(self, "_target_", target)
        object.__setattr__(self, "_handler_", handler)

    def __getattribute__(self, name):
        if object.__getattribute__(self, name):
            return object.__getattribute__(self, name)
        if self.__target__():
            return object.__getattribute__(self.__target__(), name)
        return None

    def __delattr__(self, name):
        delattr(self.__target__(), name)

    def __setattr__(self, name, value):
        setattr(self.__target__(), name, value)

    def __str__(self):
        return str(self.__target__())

    def __repr__(self):
        return repr(self.__target__())

    def __hash__(self):
        return hash(self.__target__())

    def __call__(self, *args, **kwargs):
        return self

    def __new__(cls, kinds: List[Type], handler: InvocationHandler, *args, **kwargs):
        """
        Creates an proxy instance referencing `interface`. (interface, *args, **kwargs) are
        passed to this class' __init__, so deriving classes can define an
        __init__ method of their own.
        note: _proxy_class_cache_ is unique per deriving class (each deriving
        class must hold its own cache)
        """
        proxy_class = cls.create_proxy_class(kinds)
        instance = object.__new__(proxy_class)
        instance.__init__(instance, handler)
        return instance

    _special_names_ = [
        '__abs__', '__add__', '__and__', '__cmp__', '__coerce__',
        '__contains__', '__delitem__', '__delslice__', '__div__', '__divmod__',
        '__eq__', '__float__', '__floordiv__', '__ge__', '__getitem__',
        '__getslice__', '__gt__', '__hex__', '__iadd__', '__iand__',
        '__idiv__', '__idivmod__', '__ifloordiv__', '__ilshift__', '__imod__',
        '__imul__', '__int__', '__invert__', '__ior__', '__ipow__', '__irshift__',
        '__isub__', '__iter__', '__itruediv__', '__ixor__', '__le__', '__len__',
        '__long__', '__lshift__', '__lt__', '__mod__', '__mul__', '__ne__',
        '__neg__', '__oct__', '__or__', '__pos__', '__pow__', '__radd__',
        '__rand__', '__rdiv__', '__rdivmod__', '__reduce__', '__reduce_ex__',
        '__repr__', '__reversed__', '__rfloorfiv__', '__rlshift__', '__rmod__',
        '__rmul__', '__ror__', '__rpow__', '__rrshift__', '__rshift__', '__rsub__',
        '__rtruediv__', '__rxor__', '__setitem__', '__setslice__', '__sub__',
        '__truediv__', '__xor__', 'next',
    ]

    @classmethod
    def create_proxy_class(cls, kinds: List[Type]) -> Type["Proxy"]:
        """
        Creates a proxy for the given class
        """
        qname = ",".join(sorted(map(lambda x: str(x), filter(lambda k: k != Proxy, kinds))))
        if hasattr(cls, '__proxy_classes__') and cls.__proxy_classes__.get(qname):
            return cls.__proxy_classes__.get(qname)

        if not hasattr(cls, '__proxy_classes__'):
            cls.__proxy_classes__ = {}

        namespace = {}
        for kind in kinds:
            for name in cls._special_names_:
                if hasattr(kind, name) and not hasattr(cls, name) and not namespace.get(name):
                    namespace[name] = Proxy.make_special(name)

        for kind in kinds:
            for name, method in Proxy.get_abstract_methods(kind).items():
                namespace[name] = Proxy.make_delegate(method)

        cls.__proxy_classes__[qname] = type("%s(%s)" % (cls.__name__, qname), (*kinds, cls), namespace)
        return cls.__proxy_classes__[qname]

    @staticmethod
    def new_proxy(kind: Type[T], handler: InvocationHandler) -> T:
        interface = [kind]
        if hasattr(kind, "__args__"):
            interface = kind.__args__

        return Proxy(interface, handler)

    @staticmethod
    def get_interfaces(kind: Type) -> List[Type]:
        if inspect.isabstract(kind):
            return [kind]
        interfaces = []
        if hasattr(kind, '__bases__'):
            for base in kind.__bases__:
                interfaces.append(base)
        if interfaces.__len__() < 1:
            return [kind]
        return interfaces

    @staticmethod
    def get_abstract_methods(kind: Type) -> Dict[str, Callable]:
        interfaces = Proxy.get_interfaces(kind)
        methods: Dict[str, Callable] = {}
        for interface in interfaces:
            if hasattr(interface, '__abstractmethods__'):
                for name in interface.__abstractmethods__:
                    if hasattr(interface, name):
                        methods[name] = interface.__dict__[name]

        return methods

    @staticmethod
    def make_special(method: str):
        def invoke(self, *args, **kw):
            return getattr(self.__target__(), method)(*args, **kw)

        return invoke

    @staticmethod
    def make_delegate(method):
        def invoke(self, *args, **kwargs):
            return self.__handler__().invoke(self, method, *args, **kwargs)

        return invoke



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import queue
from typing import Dict, Iterator, Any

import grpc

from mesh.grpx.marshaller import GrpcMarshaller
from mesh.macro import ServiceLoader
from mesh.mpc import Transporter, PROVIDER, MeshContext
from mesh.prsim import Context, Metadata


class GrpcBindableService(grpc.GenericRpcHandler):

    def __init__(self):
        self.marshaller = GrpcMarshaller()
        self.handlers = grpc.method_handlers_generic_handler("mesh-rpc", self.grpc_handlers())

    def grpc_handlers(self) -> Dict[str, grpc.RpcMethodHandler]:
        return {
            "v1": grpc.stream_stream_rpc_method_handler(
                self.stream_stream,
                request_deserializer=self.marshaller.deserialize,
                response_serializer=self.marshaller.serialize,
            ),
        }

    def service(self, handler_call_details):
        return self.handlers.service(handler_call_details)

    @staticmethod
    def stream_stream(iterator, ctx: grpc.ServicerContext) -> Iterator[Any]:
        return Transformer(iterator, ctx)


class Transformer(Iterator[Any]):

    def __init__(self, iterator: Iterator[Any], ctx: grpc.ServicerContext):
        self.iterator = iterator
        self.grpc_ctx = ctx
        self.transporter = ServiceLoader.load(Transporter).get(PROVIDER)
        self.mesh_ctx = self.context(ctx)

    def __next__(self) -> Any:
        buff = next(self.iterator)
        return self.transporter.transport(self.mesh_ctx, self.mesh_ctx.get_urn(), buff)

    def on_error(self, e: BaseException):
        """"""
        pass

    def on_complete(self):
        """"""
        pass

    @staticmethod
    def context(ctx: grpc.ServicerContext) -> Context:
        metadata = ctx.invocation_metadata()
        if metadata is None:
            return MeshContext.create()
        mtx = MeshContext()
        for (name, value) in metadata:
            std_name = name.replace("_", "-").lower()
            if std_name == Metadata.MESH_URN.key() or std_name == 'authority':
                mtx.urn = value
                mtx.attachments[Metadata.MESH_URN.key()] = value
                continue
            if std_name == Metadata.MESH_TRACE_ID.key():
                mtx.trace_id = value
                mtx.attachments[Metadata.MESH_TRACE_ID.key()] = value
                continue
            if std_name == Metadata.MESH_SPAN_ID.key():
                mtx.span_id = value
                mtx.attachments[Metadata.MESH_SPAN_ID.key()] = value
                continue
            mtx.attachments[std_name] = value
        return mtx



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import abstractmethod, ABC
from typing import List

from mesh.kinds import Environ, Route, Versions, Paging, Page, Institution
from mesh.macro import mpi, spi


@spi("mesh")
class Network(ABC):

    @abstractmethod
    @mpi("mesh.net.environ")
    def get_environ(self) -> Environ:
        """
        Get the meth network environment fixed information.
        :return: Fixed system information.
        """
        pass

    @abstractmethod
    @mpi("mesh.net.accessible")
    def accessible(self, route: Route) -> bool:
        """
        Check the mesh network is accessible.
        :param route: Network route.
        :return: true is accessible.
        """
        pass

    @abstractmethod
    @mpi("mesh.net.refresh")
    def refresh(self, routes: List[Route]):
        """
        Refresh the routes to mesh network.
        :param routes: Network routes.
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.net.edge")
    def get_route(self, node_id: str) -> Route:
        """
        GetNetRoute the network edge routes.
        :param node_id: node id
        :return: edge routes
        """
        pass

    @abstractmethod
    @mpi("mesh.net.edges")
    def get_routes(self) -> List[Route]:
        """
        GetNetRoute the network edge routes.
        :return: edge routes
        """
        pass

    @abstractmethod
    @mpi("mesh.net.domains")
    def get_domains(self) -> List[Route]:
        """
        GetNetDomain the network domains.
        :return: net domains
        """
        pass

    @abstractmethod
    @mpi("mesh.net.resolve")
    def put_domains(self, domains: List[Route]):
        """
        Put the network domains.
        """
        pass

    @abstractmethod
    @mpi("mesh.net.weave")
    def weave(self, route: Route) -> None:
        """
        Weave the network.
        :param route:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.net.ack")
    def ack(self, route: Route) -> None:
        """
        Acknowledge the network.
        :param route:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.net.disable")
    def disable(self, node_id: str) -> None:
        """
        Disable the network.
        :param node_id:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.net.enable")
    def enable(self, node_id: str) -> None:
        """
        Enable the network.
        :param node_id:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.net.index")
    def index(self, index: Paging) -> Page[Route]:
        """
        Index the network edges.
        :param index:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.net.version")
    def version(self, node_id: str) -> Versions:
        """
        Network environment version.
        :param node_id:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.net.instx")
    def instx(self, index: Paging) -> Page[Institution]:
        """
        Network institutions.
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.net.instr")
    def instr(self, institutions: List[Institution]) -> str:
        """
        Network institutions.
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.net.ally")
    def ally(self, node_ids: List[str]) -> None:
        """
        Network form alliance.
        """
        pass

    @abstractmethod
    @mpi("mesh.net.disband")
    def disband(self, node_ids: List[str]) -> None:
        """
        Network quit alliance.
        """
        pass



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import mesh.log as log
import unittest

import mesh.codec as codec
import mesh.mpc as mpc


class TestGrpc(unittest.TestCase):

    def test_invoke_grpc(self):
        encoder = mpc.ServiceLoader.load(codec.Codec).get('json')
        inbound = encoder.encode("")
        execution = mpc.GenericExecution()
        self.consumer = mpc.ServiceLoader.load(mpc.Consumer).get_default()
        outbound = self.consumer.consume('https://10.12.0.83:572', execution, inbound)
        log.info(outbound)


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import unittest

import mesh.codec as codec
import mesh.codec.tools as tools
import mesh.log as log
import mesh.mpc as mpc
from mesh.kinds import Paging, Page


class A:
    x: str


class TestCodec(unittest.TestCase):

    def test_json_codec(self):
        cls = Page[bytes]
        log.info(tools.get_raw_type(cls))
        paging = Paging()
        paging.index = 1
        encoder = mpc.ServiceLoader.load(codec.Codec).get(codec.Json)
        inbound = encoder.encode_string(paging)
        log.info(inbound)
        log.info(encoder.encode_string({'x': 'y'}))
        a = A()
        a.x = 'y'
        log.info(encoder.encode_string(a))

        encoder.decode_string('{}', Page[bytes])


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import List

from mesh.kinds import Event
from mesh.macro import spi
from mesh.mpc import ServiceProxy
from mesh.prsim import Publisher


@spi("mesh")
class MeshPublisher(Publisher):

    def __init__(self):
        self.proxy = ServiceProxy.default_proxy(Publisher)

    def publish(self, events: List[Event]) -> List[str]:
        return self.proxy.publish(events)

    def broadcast(self, events: List[Event]) -> List[str]:
        return self.proxy.broadcast(events)



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

"""Ring buffers for multiprocessing.
Allows multiple child Python processes started via the multiprocessing module
to read from a shared ring buffer in the parent process. For each child, a
pointer is maintained for the purpose of reading. One pointer is maintained by
for the purpose of writing. Reads may be issued in blocking or non-blocking
mode. Writes are always in non-blocking mode and will raise an exception
if the buffer is full.
"""

import contextlib
import ctypes
import multiprocessing
import struct


class Error(Exception):
    pass


class DataTooLargeError(Error, ValueError):
    pass


class WaitingForReaderError(Error):
    pass


class WaitingForWriterError(Error):
    pass


class WriterFinishedError(Error):
    pass


class AlreadyClosedError(Error):
    pass


class MustCreatedReadersBeforeWritingError(Error):
    pass


class InternalLockingError(Error):
    pass


class Position:

    def __init__(self, slot_count):
        self.counter = 0
        self.slot_count = slot_count

    @property
    def index(self):
        return self.counter % self.slot_count

    @property
    def generation(self):
        return self.counter // self.slot_count


class Pointer:

    def __init__(self, slot_count, *, start=None):
        default = start if start is not None else 0
        self.counter = multiprocessing.RawValue(ctypes.c_longlong, default)
        self.position = Position(slot_count)

    def increment(self):
        self.counter.value += 1

    def get(self):
        # Avoid reallocating Position repeatedly.
        self.position.counter = self.counter.value
        return self.position

    def set(self, counter):
        self.counter.value = counter


class RingBuffer:
    """Circular buffer class accessible to multiple threads or child processes.
    All methods are thread safe. Multiple readers and writers are permitted.
    Before kicking off multiprocessing.Process instances, first allocate all
    of the writers you'll need with new_writer() and readers with new_reader().
    Pass the Pointer value returned by the new_reader() method to the
    multiprocessing.Process constructor along with the RingBuffer instance.
    Calling  new_writer() or new_reader() from a child multiprocessing.Process
    will not work.
    """

    def __init__(self, *, slot_bytes, slot_count):
        """Initializer.
        Args:
            slot_bytes: The maximum size of slots in the buffer.
            slot_count: How many slots should be in the buffer.
        """
        self.slot_count = slot_count
        self.array = SlotArray(slot_bytes=slot_bytes, slot_count=slot_count)
        self.lock = ReadersWriterLock()
        # Each reading process may modify its own Pointer while the read
        # lock is being held. Each reading process can also load the position
        # of the writer, but not load any other readers. Each reading process
        # can also load the value of the 'active' count.
        self.readers = []
        # The writer can load and store the Pointer of all the reader Pointers
        # or the writer Pointer while the write lock is held. It can also load
        # and store the value of the 'active' acount.
        self.writer = Pointer(self.slot_count)
        self.active = multiprocessing.RawValue(ctypes.c_uint, 0)

    def new_reader(self):
        """Returns a new unique reader into the buffer.
        This must only be called in the parent process. It must not be
        called in a child multiprocessing.Process. See class docstring. To
        enforce this policy, no readers may be allocated after the first
        write has occurred.
        """
        with self.lock.for_write():
            writer_position = self.writer.get()
            if writer_position.counter > 0:
                raise MustCreatedReadersBeforeWritingError

            reader = Pointer(self.slot_count, start=writer_position.counter)
            self.readers.append(reader)
            return reader

    def new_writer(self):
        """Must be called once by each writer before any reads occur.
        Should be paired with a single subsequent call to writer_done() to
        indicate that this writer has finished and will not write any more
        data into the ring.
        """
        with self.lock.for_write():
            self.active.value += 1

    def _has_write_conflict(self, position):
        index = position.index
        generation = position.generation
        for reader in self.readers:
            # This Position and the other Position both point at the same index
            # in the ring buffer, but they have different generation numbers.
            # This means the writer can't proceed until some readers have
            # sufficiently caught up.
            reader_position = reader.get()
            if (reader_position.index == index and
                    reader_position.generation < generation):
                return True

        return False

    def try_write(self, data):
        """Tries to write the next slot, but will not block.
        Once a successful write occurs, all pending blocking_read() calls
        will be woken up to consume the newly written slot.
        Args:
            data: Bytes to write in the next available slot. Must be
                less than or equal to slot_bytes in size.
        Raises:
            WaitingForReaderError: If all of the slots are full and we need
                to wait for readers to catch up before there will be
                sufficient room to write more data. This is a sign that
                the readers can't keep up with the writer. Consider calling
                force_reader_sync() if you need to force the readers to
                catch up, but beware that means they will miss data.
        """
        with self.lock.for_write():
            if not self.active.value:
                raise AlreadyClosedError

            position = self.writer.get()
            if self._has_write_conflict(position):
                raise WaitingForReaderError

            self.array[position.index] = data
            self.writer.increment()

    def _has_read_conflict(self, reader_position):
        writer_position = self.writer.get()
        return writer_position.counter <= reader_position.counter

    def _try_read_no_lock(self, reader):
        position = reader.get()
        if self._has_read_conflict(position):
            if not self.active.value:
                raise WriterFinishedError
            else:
                raise WaitingForWriterError

        data = self.array[position.index]
        reader.increment()
        return data

    def try_read(self, reader):
        """Tries to read the next slot for a reader, but will not block.
        Args:
            reader: Position previously returned by the call to new_reader().
        Returns:
            bytearray containing a copy of the data from the slot. This
            value is mutable an can be used to back ctypes objects, NumPy
            arrays, etc.
        Raises:
            WriterFinishedError: If the RingBuffer was closed before this
                read operation began.
            WaitingForWriterError: If the given reader has already consumed
                all the data in the ring buffer and would need to block in
                order to wait for new data to arrive.
        """
        with self.lock.for_read():
            return self._try_read_no_lock(reader)

    def blocking_read(self, reader):
        """Reads the next slot for a reader, blocking if it isn't filled yet.
        Args:
            reader: Position previously returned by the call to new_reader().
        Returns:
            bytearray containing a copy of the data from the slot. This
            value is mutable an can be used to back ctypes objects, NumPy
            arrays, etc.
        Raises:
            WriterFinishedError: If the RingBuffer was closed while waiting
                to read the next operation.
        """
        with self.lock.for_read():
            while True:
                try:
                    return self._try_read_no_lock(reader)
                except WaitingForWriterError:
                    self.lock.wait_for_write()

    def force_reader_sync(self):
        """Forces all readers to skip to the position of the writer."""
        with self.lock.for_write():
            writer_position = self.writer.get()

            for reader in self.readers:
                reader.set(writer_position.counter)

            for reader in self.readers:
                p = reader.get()

    def writer_done(self):
        """Called by the writer when no more data is expected to be written.
        Should be called once for every corresponding call to new_writer().
        Once all writers have called writer_done(), a WriterFinishedError
        exception will be raised by any blocking read calls or subsequent
        calls to read.
        """
        with self.lock.for_write():
            self.active.value -= 1


class SlotArray:
    """Fast array of indexable buffers backed by shared memory.
    Assumes locking happens elsewhere.
    """

    def __init__(self, *, slot_bytes, slot_count):
        """Initializer.
        Args:
            slot_bytes: How big each buffer in the array should be.
            slot_count: How many buffers should be in the array.
        """
        self.slot_bytes = slot_bytes
        self.slot_count = slot_count
        self.length_bytes = 4
        slot_type = ctypes.c_byte * (slot_bytes + self.length_bytes)
        self.array = multiprocessing.RawArray(slot_type, slot_count)

    def __getitem__(self, i):
        data = memoryview(self.array[i])
        (length,) = struct.unpack_from('>I', data, 0)

        start = self.length_bytes
        # This must create a copy because we want the writer to be able to
        # overwrite this slot as soon as the data has been retrieved by all
        # readers. But we also want the returned bytes to be mutable so that
        # the returned data can immediately back a ctypes record using the
        # from_buffer() method (instead of from_buffer_copy()).
        return bytearray(data[start:start + length])

    def __setitem__(self, i, data):
        data_view = memoryview(data).cast('@B')
        data_size = len(data_view)
        if data_size > self.slot_bytes:
            raise DataTooLargeError('%d bytes too big for slot' % data_size)

        # Avoid copying the input data! Do only a single copy into the slot.
        slot_view = memoryview(self.array[i]).cast('@B')
        struct.pack_into('>I', slot_view, 0, data_size)
        start = self.length_bytes
        slot_view[start:start + data_size] = data_view

    def __len__(self):
        return self.slot_count


class ReadersWriterLock:
    """Multiprocessing-compatible Readers/Writer lock.
    The algorithm:
    https://en.wikipedia.org/wiki/Readers%E2%80%93writer_lock#Using_a_condition_variable_and_a_mutex
    Background on the Kernel:
    https://www.kernel.org/doc/Documentation/memory-barriers.txt
    sem_wait on Linux uses NPTL, which uses futexes:
    https://github.com/torvalds/linux/blob/master/kernel/futex.c
    Notably, futexes use the smp_mb() memory fence, which is a general write
    barrier, meaning we can assume that all memory reads and writes before
    a barrier will complete before reads and writes after the barrier, even
    if the semaphore / futex isn't actively held.
    """

    def __init__(self):
        self.lock = multiprocessing.Lock()
        self.readers_condition = multiprocessing.Condition(self.lock)
        self.writer_condition = multiprocessing.Condition(self.lock)
        self.readers = multiprocessing.RawValue(ctypes.c_uint, 0)
        self.writer = multiprocessing.RawValue(ctypes.c_bool, False)

    def _acquire_reader_lock(self):
        with self.lock:
            while self.writer.value:
                self.readers_condition.wait()

            self.readers.value += 1

    def _release_reader_lock(self):
        with self.lock:
            self.readers.value -= 1

            if self.readers.value == 0:
                self.writer_condition.notify()

    @contextlib.contextmanager
    def for_read(self):
        """Acquire the lock for reading."""
        self._acquire_reader_lock()
        try:
            yield
        finally:
            self._release_reader_lock()

    def _acquire_writer_lock(self):
        with self.lock:
            while self.writer.value or self.readers.value > 0:
                self.writer_condition.wait()

            self.writer.value = True

    def _release_writer_lock(self):
        with self.lock:
            self.writer.value = False
            self.readers_condition.notify_all()
            self.writer_condition.notify()

    @contextlib.contextmanager
    def for_write(self):
        """Acquire the lock for writing reading."""
        self._acquire_writer_lock()
        try:
            yield
        finally:
            self._release_writer_lock()

    def wait_for_write(self):
        """Block until a writer has notified readers.
        Must be called while the read lock is already held. May return
        spuriously before the writer actually did something.
        """
        with self.lock:
            if self.readers.value == 0:
                raise InternalLockingError
            self.readers.value -= 1
            self.writer_condition.notify()
            self.readers_condition.wait()
            self.readers.value += 1



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import ABC, abstractmethod
from typing import List

from mesh.kinds import Registration
from mesh.macro import mpi, spi


@spi("mesh")
class Registry(ABC):

    @abstractmethod
    @mpi("mesh.registry.put")
    def register(self, registration: Registration):
        """
        Register metadata to mesh graph database.
        :param registration:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.registry.puts")
    def registers(self, registrations: List[Registration]):
        """
        Register metadata to mesh graph database.
        :param registrations:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.registry.remove")
    def unregister(self, registration: Registration):
        """
        Unregister metadata from mesh graph database.
        :param registration:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.registry.export")
    def export(self, kind: str) -> List[Registration]:
        """
        Export register metadata of mesh graph database.
        :param kind:
        :return:
        """
        pass



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from typing import Generic

from mesh.macro import T


class Types(Generic[T]):
    pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: y.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
    name='y.proto',
    package='org.ppc.ptp',
    syntax='proto3',
    serialized_options=b'\n\013io.bfia.ptpZ\031github.com/be-io/mesh/ptp',
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x07y.proto\x12\x0borg.ppc.ptp\"\x1c\n\x0bPeekInbound\x12\r\n\x05topic\x18\x01 \x01(\t\",\n\nPopInbound\x12\r\n\x05topic\x18\x01 \x01(\t\x12\x0f\n\x07timeout\x18\x02 \x01(\x05\"\x98\x01\n\x0bPushInbound\x12\r\n\x05topic\x18\x01 \x01(\t\x12\x0f\n\x07payload\x18\x02 \x01(\x0c\x12\x38\n\x08metadata\x18\x03 \x03(\x0b\x32&.org.ppc.ptp.PushInbound.MetadataEntry\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"0\n\x0eReleaseInbound\x12\r\n\x05topic\x18\x01 \x01(\t\x12\x0f\n\x07timeout\x18\x02 \x01(\x05\"\xb4\x01\n\x11TransportOutbound\x12>\n\x08metadata\x18\x01 \x03(\x0b\x32,.org.ppc.ptp.TransportOutbound.MetadataEntry\x12\x0f\n\x07payload\x18\x02 \x01(\x0c\x12\x0c\n\x04\x63ode\x18\x03 \x01(\t\x12\x0f\n\x07message\x18\x04 \x01(\t\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x32\xa6\x02\n\x18PrivateTransferTransport\x12@\n\x04peek\x12\x18.org.ppc.ptp.PeekInbound\x1a\x1e.org.ppc.ptp.TransportOutbound\x12>\n\x03pop\x12\x17.org.ppc.ptp.PopInbound\x1a\x1e.org.ppc.ptp.TransportOutbound\x12@\n\x04push\x12\x18.org.ppc.ptp.PushInbound\x1a\x1e.org.ppc.ptp.TransportOutbound\x12\x46\n\x07release\x12\x1b.org.ppc.ptp.ReleaseInbound\x1a\x1e.org.ppc.ptp.TransportOutboundB(\n\x0bio.bfia.ptpZ\x19github.com/be-io/mesh/ptpb\x06proto3'
)




_PEEKINBOUND = _descriptor.Descriptor(
    name='PeekInbound',
    full_name='org.ppc.ptp.PeekInbound',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='topic', full_name='org.ppc.ptp.PeekInbound.topic', index=0,
            number=1, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=24,
    serialized_end=52,
)


_POPINBOUND = _descriptor.Descriptor(
    name='PopInbound',
    full_name='org.ppc.ptp.PopInbound',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='topic', full_name='org.ppc.ptp.PopInbound.topic', index=0,
            number=1, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='timeout', full_name='org.ppc.ptp.PopInbound.timeout', index=1,
            number=2, type=5, cpp_type=1, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=54,
    serialized_end=98,
)


_PUSHINBOUND_METADATAENTRY = _descriptor.Descriptor(
    name='MetadataEntry',
    full_name='org.ppc.ptp.PushInbound.MetadataEntry',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='key', full_name='org.ppc.ptp.PushInbound.MetadataEntry.key', index=0,
            number=1, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='value', full_name='org.ppc.ptp.PushInbound.MetadataEntry.value', index=1,
            number=2, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=b'8\001',
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=206,
    serialized_end=253,
)

_PUSHINBOUND = _descriptor.Descriptor(
    name='PushInbound',
    full_name='org.ppc.ptp.PushInbound',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='topic', full_name='org.ppc.ptp.PushInbound.topic', index=0,
            number=1, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='payload', full_name='org.ppc.ptp.PushInbound.payload', index=1,
            number=2, type=12, cpp_type=9, label=1,
            has_default_value=False, default_value=b"",
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='metadata', full_name='org.ppc.ptp.PushInbound.metadata', index=2,
            number=3, type=11, cpp_type=10, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    ],
    extensions=[
    ],
    nested_types=[_PUSHINBOUND_METADATAENTRY, ],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=101,
    serialized_end=253,
)


_RELEASEINBOUND = _descriptor.Descriptor(
    name='ReleaseInbound',
    full_name='org.ppc.ptp.ReleaseInbound',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='topic', full_name='org.ppc.ptp.ReleaseInbound.topic', index=0,
            number=1, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='timeout', full_name='org.ppc.ptp.ReleaseInbound.timeout', index=1,
            number=2, type=5, cpp_type=1, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=255,
    serialized_end=303,
)


_TRANSPORTOUTBOUND_METADATAENTRY = _descriptor.Descriptor(
    name='MetadataEntry',
    full_name='org.ppc.ptp.TransportOutbound.MetadataEntry',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='key', full_name='org.ppc.ptp.TransportOutbound.MetadataEntry.key', index=0,
            number=1, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='value', full_name='org.ppc.ptp.TransportOutbound.MetadataEntry.value', index=1,
            number=2, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=b'8\001',
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=206,
    serialized_end=253,
)

_TRANSPORTOUTBOUND = _descriptor.Descriptor(
    name='TransportOutbound',
    full_name='org.ppc.ptp.TransportOutbound',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='metadata', full_name='org.ppc.ptp.TransportOutbound.metadata', index=0,
            number=1, type=11, cpp_type=10, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='payload', full_name='org.ppc.ptp.TransportOutbound.payload', index=1,
            number=2, type=12, cpp_type=9, label=1,
            has_default_value=False, default_value=b"",
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='code', full_name='org.ppc.ptp.TransportOutbound.code', index=2,
            number=3, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='message', full_name='org.ppc.ptp.TransportOutbound.message', index=3,
            number=4, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    ],
    extensions=[
    ],
    nested_types=[_TRANSPORTOUTBOUND_METADATAENTRY, ],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=306,
    serialized_end=486,
)

_PUSHINBOUND_METADATAENTRY.containing_type = _PUSHINBOUND
_PUSHINBOUND.fields_by_name['metadata'].message_type = _PUSHINBOUND_METADATAENTRY
_TRANSPORTOUTBOUND_METADATAENTRY.containing_type = _TRANSPORTOUTBOUND
_TRANSPORTOUTBOUND.fields_by_name['metadata'].message_type = _TRANSPORTOUTBOUND_METADATAENTRY
DESCRIPTOR.message_types_by_name['PeekInbound'] = _PEEKINBOUND
DESCRIPTOR.message_types_by_name['PopInbound'] = _POPINBOUND
DESCRIPTOR.message_types_by_name['PushInbound'] = _PUSHINBOUND
DESCRIPTOR.message_types_by_name['ReleaseInbound'] = _RELEASEINBOUND
DESCRIPTOR.message_types_by_name['TransportOutbound'] = _TRANSPORTOUTBOUND
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PeekInbound = _reflection.GeneratedProtocolMessageType('PeekInbound', (_message.Message,), {
    'DESCRIPTOR' : _PEEKINBOUND,
    '__module__' : 'y_pb2'
    # @@protoc_insertion_point(class_scope:org.ppc.ptp.PeekInbound)
})
_sym_db.RegisterMessage(PeekInbound)

PopInbound = _reflection.GeneratedProtocolMessageType('PopInbound', (_message.Message,), {
    'DESCRIPTOR' : _POPINBOUND,
    '__module__' : 'y_pb2'
    # @@protoc_insertion_point(class_scope:org.ppc.ptp.PopInbound)
})
_sym_db.RegisterMessage(PopInbound)

PushInbound = _reflection.GeneratedProtocolMessageType('PushInbound', (_message.Message,), {

    'MetadataEntry' : _reflection.GeneratedProtocolMessageType('MetadataEntry', (_message.Message,), {
        'DESCRIPTOR' : _PUSHINBOUND_METADATAENTRY,
        '__module__' : 'y_pb2'
        # @@protoc_insertion_point(class_scope:org.ppc.ptp.PushInbound.MetadataEntry)
    })
    ,
    'DESCRIPTOR' : _PUSHINBOUND,
    '__module__' : 'y_pb2'
    # @@protoc_insertion_point(class_scope:org.ppc.ptp.PushInbound)
})
_sym_db.RegisterMessage(PushInbound)
_sym_db.RegisterMessage(PushInbound.MetadataEntry)

ReleaseInbound = _reflection.GeneratedProtocolMessageType('ReleaseInbound', (_message.Message,), {
    'DESCRIPTOR' : _RELEASEINBOUND,
    '__module__' : 'y_pb2'
    # @@protoc_insertion_point(class_scope:org.ppc.ptp.ReleaseInbound)
})
_sym_db.RegisterMessage(ReleaseInbound)

TransportOutbound = _reflection.GeneratedProtocolMessageType('TransportOutbound', (_message.Message,), {

    'MetadataEntry' : _reflection.GeneratedProtocolMessageType('MetadataEntry', (_message.Message,), {
        'DESCRIPTOR' : _TRANSPORTOUTBOUND_METADATAENTRY,
        '__module__' : 'y_pb2'
        # @@protoc_insertion_point(class_scope:org.ppc.ptp.TransportOutbound.MetadataEntry)
    })
    ,
    'DESCRIPTOR' : _TRANSPORTOUTBOUND,
    '__module__' : 'y_pb2'
    # @@protoc_insertion_point(class_scope:org.ppc.ptp.TransportOutbound)
})
_sym_db.RegisterMessage(TransportOutbound)
_sym_db.RegisterMessage(TransportOutbound.MetadataEntry)


DESCRIPTOR._options = None
_PUSHINBOUND_METADATAENTRY._options = None
_TRANSPORTOUTBOUND_METADATAENTRY._options = None

_PRIVATETRANSFERTRANSPORT = _descriptor.ServiceDescriptor(
    name='PrivateTransferTransport',
    full_name='org.ppc.ptp.PrivateTransferTransport',
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=489,
    serialized_end=783,
    methods=[
        _descriptor.MethodDescriptor(
            name='peek',
            full_name='org.ppc.ptp.PrivateTransferTransport.peek',
            index=0,
            containing_service=None,
            input_type=_PEEKINBOUND,
            output_type=_TRANSPORTOUTBOUND,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name='pop',
            full_name='org.ppc.ptp.PrivateTransferTransport.pop',
            index=1,
            containing_service=None,
            input_type=_POPINBOUND,
            output_type=_TRANSPORTOUTBOUND,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name='push',
            full_name='org.ppc.ptp.PrivateTransferTransport.push',
            index=2,
            containing_service=None,
            input_type=_PUSHINBOUND,
            output_type=_TRANSPORTOUTBOUND,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name='release',
            full_name='org.ppc.ptp.PrivateTransferTransport.release',
            index=3,
            containing_service=None,
            input_type=_RELEASEINBOUND,
            output_type=_TRANSPORTOUTBOUND,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ])
_sym_db.RegisterServiceDescriptor(_PRIVATETRANSFERTRANSPORT)

DESCRIPTOR.services_by_name['PrivateTransferTransport'] = _PRIVATETRANSFERTRANSPORT

# @@protoc_insertion_point(module_scope)



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import abstractmethod
from typing import Any, Type

from mesh.cause import MeshException
from mesh.macro import spi, T


@spi("json")
class Codec:

    @abstractmethod
    def encode(self, value: T) -> bytes:
        pass

    @abstractmethod
    def decode(self, value: bytes, kind: Type[T]) -> T:
        pass

    def encode_string(self, value: Any) -> str:
        return self.encode(value).decode('UTF-8')

    def decode_string(self, value: str, kind: Type[T]) -> T:
        return self.decode(value.encode('UTF-8'), kind)


class CodecError(MeshException):
    def __init__(self, message: str):
        super().__init__("E000001", message)


